"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GMail = exports.Mail = void 0;
const Mailer = require("nodemailer");
const os_1 = require("os");
const Log_1 = require("./Log");
const { google } = require('googleapis');
const MailComposer = require('nodemailer/lib/mail-composer');
class Mail {
    constructor(_config) {
        this._config = _config;
        this._template = '<p><!-- body --></p><p>At: <!-- timeStamp --></p>';
        if (!this._config.smtp)
            this._config.smtp = { disabled: true };
        if (this._config.smtp.disabled === true)
            return; // don't analyze config if disabled
        if (!this._config.smtp)
            throw new Error(`[smtp] not set`);
        if (!this._config.smtp.host)
            throw new Error(`[smtp.host] not set`);
        if (!this._config.smtp)
            throw new Error(`[smtp.port] not set`);
    }
    configChanged() {
        if (!this._config.mailTo)
            throw new Error(`[mailTo] not set`);
    }
    async send(message) {
        if (this._config.smtp.disabled === true) {
            Log_1.debug('mail sending is disbled in config');
            return;
        }
        const temp = {
            host: this._config.smtp.host,
            port: this._config.smtp.port,
            tls: { rejectUnauthorized: false },
            secure: this._config.smtp.secure === true,
            auth: null
        };
        if (this._config.smtp.user)
            temp.auth = {
                user: this._config.smtp.user,
                pass: this._config.smtp.password
            };
        const transport = Mailer.createTransport(temp), headers = {};
        if (message.priority)
            headers['importance'] = message.priority;
        await transport.sendMail({
            to: this._config.mailTo,
            from: this._config.smtp.from || this._config.smtp.user,
            replyTo: this._config.replyTo,
            subject: `pm2-health-custom: ${os_1.hostname()}, ${message.subject}`,
            html: this._template
                .replace(/<!--\s*body\s*-->/, message.body)
                .replace(/<!--\s*timeStamp\s*-->/, new Date().toISOString()),
            attachments: message.attachements,
            headers
        });
    }
}
exports.Mail = Mail;
class GMail extends Mail {
    constructor(_config) {
        super(_config);
        this.TOKEN_PATH = 'token.json';
        if (!this._config.gmail)
            this._config.gmail = { disabled: true };
        if (this._config.gmail.disabled === true)
            return; // don't analyze config if disabled
        if (!this._config.gmail)
            throw new Error(`[gmail] not set`);
        if (!this._config.gmail.type)
            throw new Error(`[smtp.type] not set`);
        if (!this._config.gmail.clientId)
            throw new Error(`[smtp.clientId] not set`);
        if (!this._config.gmail.clientSecret)
            throw new Error(`[smtp.clientSecret] not set`);
        if (!this._config.gmail.user)
            throw new Error(`[smtp.user] not set`);
        if (!this._config.gmail.refreshToken)
            throw new Error(`[smtp.refreshToken] not set`);
    }
    async send(message) {
        const oAuth2Client = new google.auth.OAuth2(this._config.gmail.clientId, this._config.gmail.clientSecret, this._config.gmail.redirect_uris[0]);
        const token = {
            refresh_token: this._config.gmail.refreshToken
        };
        oAuth2Client.setCredentials(token);
        const accessToken = await oAuth2Client.getAccessToken();
        oAuth2Client.setCredentials(accessToken.res.data);
        const headers = {};
        if (message.priority)
            headers['importance'] = message.priority;
        const mail = new MailComposer({
            to: this._config.mailTo,
            from: this._config.gmail.from,
            replyTo: this._config.replyTo,
            subject: `pm2-health-custom: ${os_1.hostname()}, ${message.subject}`,
            html: this._template
                .replace(/<!--\s*body\s*-->/, message.body)
                .replace(/<!--\s*timeStamp\s*-->/, new Date().toISOString()),
            attachments: message.attachements,
            headers
        });
        console.log(this._template);
        mail.compile().build((error, msg) => {
            if (error)
                return console.log('Error compiling email ' + error);
            const encodedMessage = Buffer.from(msg)
                .toString('base64')
                .replace(/\+/g, '-')
                .replace(/\//g, '_')
                .replace(/=+$/, '');
            const gmail = google.gmail({ version: 'v1', auth: oAuth2Client });
            gmail.users.messages.send({
                userId: 'me',
                resource: {
                    raw: encodedMessage
                }
            }, (err, result) => {
                if (err)
                    return console.log('NODEMAILER - The API returned an error: ' + err);
                console.log('NODEMAILER - Sending email reply from server:', result.data);
            });
        });
    }
}
exports.GMail = GMail;
//# sourceMappingURL=Mail.js.map